#include "sysfuncs.h"
#include "netorder.h"
#include "messages.h"
#include "memlib.h"
#include "local_strings.h"

#include "backend.h"
#include "StemBlock.h"
#include "mg_errors.h"

/**
 * GRB: The abort macro defined below is used to provide relatively
 *      graceful failure when an error occurs - it calls the Recovery
 *      label defined after the standard return from the fn to give
 *      cascading error handling (reducing repeated code).  The number
 *      given in any given abort() "call" decides where in the error 
 *      handling cascade the recovery starts 
 */

#define abort(x) abortvar=x;goto Recovery;

void StemBlock_ReadWordString(u_char **base, u_char *prev)
{
  mgString::readString(base, prev);
  /*
  unsigned copy, suff;

  copy  = **base;
  *base = *base + 1;
  suff  = **base;
  *base = *base + 1;

  bcopy ((void *) *base, (void *) (prev + copy + 1), suff);
  *prev = copy + suff;
  *base += suff;
  */
}

stemmed_dict *
ReadStemDictBlk (File * stem_file, StemBlock_errMethod method)
{
  unsigned long i;
  stemmed_dict *sd;
  u_char *buffer;
  int abortvar;

  if (!(sd = new stemmed_dict))
    {
      abort(0);
    }

  sd->stem_file = stem_file;
  sd->MemForStemDict = 0;

  Fread (&sd->sdh, sizeof (sd->sdh), 1, stem_file);
  /* [RPAP - Jan 97: Endian Ordering] */
  NTOHUL(sd->sdh.lookback);
  NTOHUL(sd->sdh.block_size);
  NTOHUL(sd->sdh.num_blocks);
  NTOHUL(sd->sdh.blocks_start);
  NTOHUL(sd->sdh.index_chars);
  NTOHUL(sd->sdh.num_of_docs);
  NTOHUL(sd->sdh.static_num_of_docs);
  NTOHUL(sd->sdh.num_of_words);
  NTOHUL(sd->sdh.stem_method);
  NTOHUL(sd->sdh.indexed);

  if (!(buffer = new u_char[sd->sdh.index_chars]))
    {
      abort(1);
    }
  sd->MemForStemDict += sd->sdh.index_chars;

  if (!(sd->index = new (u_char *)[sd->sdh.num_blocks]))
    {
      abort(2);
    }
  sd->MemForStemDict += sd->sdh.num_blocks * sizeof (*sd->index);

  if (!(sd->pos = new (long unsigned int)[sd->sdh.num_blocks]))
    {
      abort(3);
    }
  sd->MemForStemDict += sd->sdh.num_blocks * sizeof (*sd->pos);

  if (!(sd->buffer = new u_char[sd->sdh.block_size]))
    {
      abort(4);
    }
  sd->MemForStemDict += sd->sdh.block_size * sizeof (*sd->buffer);

  sd->active = -1;

  for (i = 0; i < sd->sdh.num_blocks; i++)
    {
      register u_char len;
      // Read an MG string from the blocked dictionary
      sd->index[i] = buffer;
      len = Getc (stem_file);
      *buffer++ = len;
      Fread (buffer, sizeof (u_char), len, stem_file);
      buffer += len;

      // Read the pointer into the blocks
      Fread (&sd->pos[i], sizeof (*sd->pos), 1, stem_file);

      NTOHUL(sd->pos[i]); /* [RPAP - Jan 97: Endian Ordering] */
    }

  mg_errno = MG_NOERROR;

  /*  fprintf (stderr, "mem for stem dict = %i\n", sd->MemForStemDict); */

  return sd;

 Recovery:
  switch (abortvar)
    {
      case 4:
        delete sd->pos;
      case 3:
        delete sd->index;
      case 2:
        delete buffer;
      case 1:
        delete sd;
      break;

    }

  switch (method)
    {
      case _MGErr:
      {
	mg_errno = MG_NOMEM;
      }
      break;

      case _FatalErr:
      {
	FatalError(1, "Could not allocate memory for stemmed dict");
      }
      break;
    }
  return NULL;
}


void StemBlock_read(stemmed_dict *sd,
		    int block,
		    unsigned long **first_word,
		    unsigned long **last_invf_len,
		    unsigned short **num_words,
		    unsigned short **index,
		    int *num_indexes)
{
      int i;

      Fseek (sd->stem_file, sd->pos[block] + sd->sdh.blocks_start, 0);
      Fread (sd->buffer, sd->sdh.block_size, sizeof (u_char), sd->stem_file);
      sd->active = sd->pos[block];

      /* [RPAP - Jan 97: Endian Ordering] */
      // first_word is the word number of the first word in the block
      *first_word = (unsigned long *) (sd->buffer);
      NTOHUL(**first_word);

      *last_invf_len = (unsigned long *) ((*first_word) + 1);
      NTOHUL(**last_invf_len);

      // num_words is the number of words in the block
      *num_words = (unsigned short *) ((*last_invf_len) + 1);
      NTOHUS(**num_words);

      *index = (*num_words) + 1;
      *num_indexes = (((**num_words) - 1) / sd->sdh.lookback) + 1;

      for (i = 0; i < *num_indexes; i++)
	NTOHUS((*index)[i]);
}

